import logging

import great_expectations as gx
from pyspark.sql import DataFrame
from pyspark.sql.types import *

from .consts import project_root_dir
from .validation import Validator


class FabricDataGuard:
    def __init__(
        self,
        datasource_name: str,
        data_asset_name: str,
        project_root_dir: str = project_root_dir,
    ):
        # User parameters to be provided
        self.project_root_dir = project_root_dir
        self.datasource_name = datasource_name
        self.data_asset_name = data_asset_name
        self.suite_name = f"{self.data_asset_name}Suite"

        # Initialize great expectation objects
        self.context = gx.get_context(
            mode="file", project_root_dir=self.project_root_dir
        )
        self.datasource = self._set_datasource(self.datasource_name)
        self.data_asset = self._set_data_asset(self.data_asset_name)
        self.expectation_suite = self._create_expectation_suite(self.suite_name)
        self.batch_definition = self._create_batch_definition()
        self.validation_definition = self._create_validation_definition()

    def _set_datasource(self, datasource_name: str):
        datasource = (
            self.context.data_sources.add_or_update_spark(name=datasource_name)
            if not any(ds == datasource_name for ds in self.context.data_sources.all())
            else self.context.data_sources.get(datasource_name)
        )
        return datasource

    def _set_data_asset(self, data_asset_name: str):
        try:
            data_asset = self.context.data_sources.get(self.datasource_name).get_asset(
                data_asset_name
            )
        except:
            data_asset = self.datasource.add_dataframe_asset(name=data_asset_name)
        return data_asset

    def _create_expectation_suite(self, suite_name: str):
        expectation_suite = (
            self.context.suites.add(gx.ExpectationSuite(name=suite_name))
            if not any(cs["name"] == suite_name for cs in self.context.suites.all())
            else self.context.suites.get(suite_name)
        )
        return expectation_suite

    def _create_batch_definition(self):
        batch_definition_name = f"{self.data_asset_name}BatchDefinition"
        try:
            batch_definition = self.data_asset.get_batch_definition(
                batch_definition_name
            )
        except:
            batch_definition = self.data_asset.add_batch_definition_whole_dataframe(
                batch_definition_name
            )
        return batch_definition

    def _create_validation_definition(self):
        definition_name = f"{self.data_asset_name}ValidationDefinition"
        validation_definition = (
            self.context.validation_definitions.add(
                gx.ValidationDefinition(
                    data=self.batch_definition,
                    suite=self.expectation_suite,
                    name=definition_name,
                )
            )
            if not any(
                cs.name == definition_name
                for cs in self.context.validation_definitions.all()
            )
            else self.context.validation_definitions.get(definition_name)
        )
        return validation_definition

    def add_expectation(self, expectations):
        """
        Add one or more expectations to the expectation suite.

        Args:
            expectations: A single expectation or a list of expectations to add.
        """
        if not isinstance(expectations, list):
            expectations = [expectations]

        for expectation in expectations:
            try:
                self.expectation_suite.add_expectation(expectation)
                logging.info(f"Added expectation: {expectation}")
            except Exception as e:
                logging.info(f"Expectation already exists in the suite. Updating it.")
                self.expectation_suite.remove_expectation(
                    expectation.configuration, remove_multiple_matches=True
                )
                self.expectation_suite.add_expectation(expectation)

    def run_validation(self, dataframe: DataFrame, **kwargs):
        validator = Validator(self)
        return validator.validate(dataframe, **kwargs)



import os
from datetime import datetime
from pathlib import Path

from IPython.display import HTML, display


def get_latest_html_file(root_folder):
    """
    Retrieve the path of the HTML file in the latest subfolder based on the subfolder name.

    Args:
    root_folder (str): The path to the root folder containing subfolders named with datetime values.

    Returns:
    str: The full path to the HTML file in the latest subfolder, or None if no HTML file is found.
    """
    latest_subfolder = None
    latest_date = datetime.min
    for subfolder_name in os.listdir(root_folder):
        subfolder_path = Path(root_folder) / subfolder_name
        if subfolder_path.is_dir() and subfolder_name.endswith("Z"):
            try:
                folder_date = datetime.strptime(subfolder_name, "%Y%m%dT%H%M%S.%fZ")
                if folder_date > latest_date:
                    latest_date = folder_date
                    latest_subfolder = subfolder_path
            except ValueError:
                continue
    if latest_subfolder:
        html_files = list(latest_subfolder.glob("*.html"))
        if html_files:
            return str(html_files[0])
    return None


def show_great_expectations_html(project_root_dir, suite_name):
    """
    Display the HTML content of the latest Great Expectations validation result.

    Args:
    project_root_dir (str): The root directory of the Great Expectations project.
    suite_name (str): The name of the expectation suite.
    """
    root_folder = f"{project_root_dir}/gx/uncommitted/data_docs/local_site/validations/{suite_name}/{suite_name}_run"
    file_path = get_latest_html_file(root_folder)
    if file_path:
        with open(file_path, "r") as f:
            html_content = f.read()
        display(HTML(html_content))
    else:
        print("No HTML file found for the latest validation.")



import great_expectations as gx
from pyspark.sql.types import *


def parse_validation_results(validation_results):
    """
    Parses validation results and creates a PySpark DataFrame with extracted attributes and details
    of individual test results.
    Args:
        - validation_results (dict): A dictionary containing validation results, where each entry includes
        metadata, statistics, and test outcomes.

    Returns:
        pyspark.sql.DataFrame: A PySpark DataFrame with parsed validation results. The main attributes
        are stored as separate columns in Pascal Case
    """

    parsed_data = []
    for validation_name, v in validation_results.items():
        meta = v.get("meta", {})
        statistics = v.get("statistics", {})
        active_batch_definition = meta.get("active_batch_definition", {})
        run_id = meta.get("run_id", gx.core.RunIdentifier())

        run_time = run_id.run_time
        validation_time = meta.get("validation_time")

        main_row = {
            "DatasourceName": active_batch_definition.get("datasource_name"),
            "DataAssetName": active_batch_definition.get("data_asset_name"),
            "ValidationId": meta.get("validation_id"),
            "CheckpointId": meta.get("checkpoint_id"),
            "RunName": run_id.run_name,
            "RunTime": run_time.isoformat() if run_time else None,
            "ValidationTime": validation_time.isoformat() if validation_time else None,
            "TestStatus": "Success" if v.get("success", False) else "Failure",
            "SuiteName": v.get("suite_name"),
            "EvaluatedExpectations": statistics.get("evaluated_expectations"),
            "SuccessfulExpectations": statistics.get("successful_expectations"),
            "UnsuccessfulExpectations": statistics.get("unsuccessful_expectations"),
            "SuccessPercent": statistics.get("success_percent"),
        }

        succeeded_detail_results = []
        failed_detail_results = []
        for result in v.get("results", []):
            expectation_config = result.get("expectation_config", {})
            result_data = result.get("result", {})

            detail_result = {
                "TestId": expectation_config.get("id"),
                "ColumnName": expectation_config.get("kwargs", {}).get("column"),
                "TestStatus": "Success" if result.get("success", False) else "Failure",
                "ElementCount": result_data.get("element_count"),
                "UnexpectedCount": result_data.get("unexpected_count"),
                "UnexpectedPercent": result_data.get("unexpected_percent"),
                "ExpectationType": expectation_config.get("type"),
                "UnexpectedIndexQuery": result_data.get("unexpected_index_query"),
            }

            if result.get("success", False):
                succeeded_detail_results.append(
                    {
                        "TestId": detail_result["TestId"],
                        "ColumnName": detail_result["ColumnName"],
                        "TestStatus": detail_result["TestStatus"],
                        "ExpectationType": detail_result["ExpectationType"],
                    }
                )
            else:
                unexpected_index_list = result_data.get(
                    "partial_unexpected_index_list", []
                )
                detail_result["UnexpectedIndexList"] = [
                    {str(k): str(v) for k, v in item.items()}
                    for item in unexpected_index_list
                ]
                failed_detail_results.append(detail_result)

        main_row["SucceededDetailResults"] = succeeded_detail_results
        main_row["FailedDetailResults"] = failed_detail_results
        parsed_data.append(main_row)

    return parsed_data



import sempy
import sempy.fabric as fabric
from pyspark.sql import DataFrame, SparkSession

spark = SparkSession.builder.getOrCreate()


def get_lakehouse_name():
    """
    Function to retrieve the lakehouse name attached to notebook and workspace in Fabric
    """
    try:
        lakehouse_name = spark.conf.get("trident.lakehouse.name")
    except:
        lakehouse_name = fabric.resolve_item_name(fabric.get_lakehouse_id())
    return lakehouse_name


def get_workspace_name():
    """
    Function to retrieve the workspace name of notebook running data quality checks
    """
    try:
        workspace_name = fabric.resolve_workspace_name()
    except:
        workspace_name = mssparkutils.env.getWorkspaceName()
    return workspace_name


def build_table_path(table_name, workspace_name=None, lakehouse_name=None):
    """
    Build the path where to save the logs table in the Lakehouse.

    Args:
        table_name (str): Name of the great expectation log result table.
        workspace_name (str, optional): Name of the workspace. Defaults to None, in which case it is retrieved internally.
        lakehouse_name (str, optional): Name of the lakehouse. Defaults to None, in which case it is retrieved internally.

    Returns:
        str: Full path where to store the table in the lakehouse.
    """
    if lakehouse_name is None:
        lakehouse_name = get_lakehouse_name()
    if workspace_name is None:
        workspace_name = get_workspace_name()
    return f"abfss://{workspace_name}@onelake.dfs.fabric.microsoft.com/{lakehouse_name}.Lakehouse/Tables/{table_name}"


def append_logs_to_table(
    df: DataFrame,
    table_name: str,
    workspace_name: str = None,
    lakehouse_name: str = None,
):
    """
    Function to add the results of great expectations tests in a table in the lake house.
    If a lakehouse is not provided, it creates the table in the lakehouse attached to the notebook

    Args:
        df (DataFrame): Dataframe to be saved in the lakehouse
        table_name (str): The name of the target table
        workspace_name (str, optional): Name of the workspace. Defaults to None, in which case it is retrieved internally.
        lakehouse_name (str, optional): Name of the lakehouse. Defaults to None, in which case it is retrieved internally.
    """
    target_path = build_table_path(table_name, workspace_name, lakehouse_name)
    df.write.mode("append").format("delta").option("overwriteSchema", "true").save(
        target_path
    )


def create_dataframe(parsed_data, main_schema):
    """
    Create a dataframe from the parsed results of great expectation output
    """
    df = spark.createDataFrame(parsed_data, schema=main_schema)
    return df



import great_expectations as gx


def create_expectation(expectation_type, **kwargs):
    """
    A simple wrapper to create Great Expectations expectation objects.
    This function exists to provide a consistent interface and potential for future extensions.
    Args:
        expectation_type (str): The type of expectation to create.
        **kwargs: Arguments to pass to the expectation constructor.

    Returns:
        gx.Expectation: The created expectation object
    """
    expectation_class = getattr(gx.expectations, expectation_type)
    return expectation_class(**kwargs)



import great_expectations as gx


def create_checkpoint(fabric_data_guard, **kwargs):
    """
    Creates a checkpoint for validations with optional notifications.

    Args:
        fabric_data_guard (FabricDataGuard): The FabricDataGuard instance.
        slack_notification (bool): Whether to enable Slack notifications.
        email_notification (bool): Whether to enable email notifications.
        teams_notification (bool): Whether to enable Microsoft Teams notifications.
        **kwargs: Additional keyword arguments for customizing notifications.

    Returns:
        gx.Checkpoint: The created or retrieved checkpoint.
    """
    checkpoint_name = f"{fabric_data_guard.datasource_name}AnalysisCheckpoint"

    # Base actions for the checkpoint
    actions = [gx.checkpoint.UpdateDataDocsAction(name="update_all_data_docs")]

    # Add Slack notification action if enabled
    if kwargs.get("slack_notification", False):
        actions.append(
            gx.checkpoint.SlackNotificationAction(
                name="send_slack_notification_on_failed_expectations",
                slack_token=kwargs.get("slack_token"),
                slack_channel=kwargs.get("slack_channel"),
                notify_on=kwargs.get("notify_on", "failure"),
                show_failed_expectations=kwargs.get("show_failed_expectations", True),
            )
        )

    # Add Email notification action if enabled
    if kwargs.get("email_notification", False):
        actions.append(
            gx.checkpoint.EmailAction(
                name="send_email_notification_on_validation_result",
                smtp_address=kwargs.get("smtp_address", "smtp.example.com"),
                smtp_port=kwargs.get("smtp_port", 587),
                sender_login=kwargs.get("sender_login", "your_email@example.com"),
                sender_password=kwargs.get("sender_password", "your_password"),
                receiver_emails=kwargs.get("receiver_emails", "recipient@example.com"),
                use_tls=kwargs.get("use_tls", True),
                notify_on=kwargs.get("notify_on", "failure"),
            )
        )

    # Add Microsoft Teams notification action if enabled
    if kwargs.get("teams_notification", False):
        actions.append(
            gx.checkpoint.MicrosoftTeamsNotificationAction(
                name="send_microsoft_teams_notification_on_validation_result",
                teams_webhook=kwargs.get("teams_webhook"),
                notify_on=kwargs.get("notify_on", "failure"),
            )
        )

    # Create or retrieve the checkpoint with the defined actions
    checkpoint = (
        fabric_data_guard.context.checkpoints.add(
            gx.Checkpoint(
                name=checkpoint_name,
                validation_definitions=[fabric_data_guard.validation_definition],
                actions=actions,
                result_format={
                    "result_format": "COMPLETE",
                    "unexpected_index_column_names": kwargs.get(
                        "unexpected_identifiers"
                    ),
                },
            )
        )
        if not any(
            cs.name == checkpoint_name
            for cs in fabric_data_guard.context.checkpoints.all()
        )
        else fabric_data_guard.context.checkpoints.get(checkpoint_name)
    )

    return checkpoint



import great_expectations as gx
import pyspark.sql.functions as F
from pyspark.sql import DataFrame
from pyspark.sql.types import *

from .checkpoint import create_checkpoint
from .consts import main_schema
from .display import show_great_expectations_html
from .result_parser import parse_validation_results
from .utils import append_logs_to_table, create_dataframe


class Validator:
    def __init__(self, fabric_data_guard):
        self.fdg = fabric_data_guard

    def validate(
        self, dataframe: DataFrame, checkpoint: gx.Checkpoint = None, **kwargs
    ):
        """
        Runs validation on the provided dataframe using the configured expectations.

        Args:
            dataframe (DataFrame): The Spark DataFrame to validate.
            checkpoint (gx.Checkpoint, optional): A pre-configured checkpoint. If None, a new one will be created.
            **kwargs: Additional arguments to pass to the checkpoint run.

        Returns:
            dict: The validation results.
        """
        if not all(
            [
                self.fdg.datasource_name,
                self.fdg.data_asset_name,
                self.fdg.expectation_suite,
                self.fdg.batch_definition,
            ]
        ):
            raise ValueError(
                "Datasource, data asset, expectation suite, and batch definition must be set before running validation"
            )

        if checkpoint is None:
            checkpoint = create_checkpoint(self.fdg, **kwargs)

        batch_parameters = {"dataframe": dataframe}

        run_id = gx.core.RunIdentifier(
            run_name=f"{self.fdg.expectation_suite.name}_run"
        )
        results = checkpoint.run(run_id=run_id, batch_parameters=batch_parameters)

        table_name = kwargs.get(
            "table_name", f"{self.fdg.datasource_name}DataQualityResults"
        )
        workspace_name = kwargs.get("workspace_name", None)
        lakehouse_name = kwargs.get("lakehouse_name", None)
        display_html = kwargs.get("display_html", True)

        parsed_results = self._process_results(
            results.run_results,
            table_name,
            workspace_name,
            lakehouse_name,
            display_html,
        )

        return parsed_results

    def _process_results(
        self, results, table_name, workspace_name, lakehouse_name, display_html
    ):
        """
        Processes the validation results by parsing them, logging to a table, and displaying the HTML output.

        Args:
            results (dict): The validation results from the checkpoint run.
            table_name (str): Great expectations log results table name
            workspace_name (str) : The name of the workspace where the log table will be saved
            lakehouse_name (str) : The name of the lakehouse containing the log results table
            display_html (bool) : Whether or not to display the HTNL after validation is done (default is True)
        """
        parsed_results = parse_validation_results(results)
        df_result = create_dataframe(parsed_results, main_schema)
        for column in ["RunTime", "ValidationTime"]:
            df_result = df_result.withColumn(
                column, F.col(column).cast(TimestampType())
            )

        append_logs_to_table(df_result, table_name, workspace_name, lakehouse_name)

        if display_html:
            show_great_expectations_html(
                self.fdg.project_root_dir, self.fdg.expectation_suite.name
            )

        return parsed_results



from pyspark.sql.types import *

# Define the schema for the SucceededDetailResults struct
succeeded_detail_results_schema = StructType(
    [
        StructField("TestId", StringType(), True),
        StructField("ColumnName", StringType(), True),
        StructField("TestStatus", StringType(), True),
        StructField("ExpectationType", StringType(), True),
    ]
)

# Define the schema for the FailedDetailResults struct
failed_detail_results_schema = StructType(
    [
        StructField("TestId", StringType(), True),
        StructField("ColumnName", StringType(), True),
        StructField("TestStatus", StringType(), True),
        StructField("ElementCount", IntegerType(), True),
        StructField("UnexpectedCount", IntegerType(), True),
        StructField("UnexpectedPercent", FloatType(), True),
        StructField("ExpectationType", StringType(), True),
        StructField("UnexpectedIndexQuery", StringType(), True),
        StructField(
            "UnexpectedIndexList", ArrayType(MapType(StringType(), StringType())), True
        ),
    ]
)

# Define the schema for the main DataFrame
main_schema = StructType(
    [
        StructField("DatasourceName", StringType(), True),
        StructField("DataAssetName", StringType(), True),
        StructField("ValidationId", StringType(), True),
        StructField("CheckpointId", StringType(), True),
        StructField("RunName", StringType(), True),
        StructField("RunTime", StringType(), True),
        StructField("ValidationTime", StringType(), True),
        StructField("TestStatus", StringType(), True),
        StructField("SuiteName", StringType(), True),
        StructField("EvaluatedExpectations", IntegerType(), True),
        StructField("SuccessfulExpectations", IntegerType(), True),
        StructField("UnsuccessfulExpectations", IntegerType(), True),
        StructField("SuccessPercent", FloatType(), True),
        StructField(
            "SucceededDetailResults", ArrayType(succeeded_detail_results_schema), True
        ),
        StructField(
            "FailedDetailResults", ArrayType(failed_detail_results_schema), True
        ),
    ]
)

project_root_dir = "/lakehouse/default/Files"
